﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;
using HeuristicLab.Common;

namespace HeuristicLab.Algorithms.Bandits {
  // for test case 1 in Extreme Bandits paper (Carpentier, NIPS 2014)
  public class ParetoBandit : IBandit {
    private double[] alpha;
    private double[] pZero;
    public int NumArms { get { return alpha.Length; } }
    public double OptimalExpectedReward { get; private set; } // reward of the best arm, for calculating regret
    public int OptimalExpectedRewardArm { get; private set; }
    public int OptimalMaximalRewardArm { get; private set; }
    public double MaxReward { get; private set; }
    public double MinReward { get; private set; }
    private readonly Random random;
    public ParetoBandit(Random random, IEnumerable<double> alpha) : this(random, alpha, alpha.Select(_ => 0.0)) { }
    public ParetoBandit(Random random, IEnumerable<double> alpha, IEnumerable<double> pZero) { // probability of a zero reward
      this.alpha = alpha.ToArray();
      this.pZero = pZero.ToArray();
      this.random = random;

      // find optimal arms using empirical estimates
      var bestExpReward = double.NegativeInfinity;
      var bestMaxReward = double.NegativeInfinity;
      for (int k = 0; k < NumArms; k++) {
        double expReward = 0.0;
        double maxReward = double.NegativeInfinity;
        for (int i = 0; i < 100000; i++) {
          var r = Pull(k);
          expReward += r;
          maxReward = Math.Max(maxReward, r);
        }
        expReward /= 100000;

        if (expReward > bestExpReward) {
          bestExpReward = expReward;
          OptimalExpectedRewardArm = k;
          OptimalExpectedReward = expReward;
        }
        if (maxReward > bestMaxReward) {
          bestMaxReward = maxReward;
          OptimalMaximalRewardArm = k;
        }
      }
    }

    public double Pull(int arm) {
      if (random.NextDouble() < pZero[arm]) return 0.0;
      var u = random.NextDouble();
      return Math.Pow(1.0 - u, (-1 / alpha[arm]));
    }
  }
}
